@@ -1,15 +1,17 @@
 """Transformers for numerical data."""
-
 import copy
+import sys
 
 import numpy as np
 import pandas as pd
 import scipy
-from copulas import EPSILON, univariate
 
 from rdt.transformers.base import BaseTransformer
 from rdt.transformers.null import NullTransformer
 
+EPSILON = np.finfo(np.float32).eps
+MAX_DECIMALS = sys.float_info.dig - 1
+
 
 class NumericalTransformer(BaseTransformer):
     """Transformer for numerical data.
@@ -35,15 +37,59 @@ class NumericalTransformer(BaseTransformer):
             If ``True``, always create the new column whether there are null values or not.
             If ``False``, do not create the new column.
             Defaults to ``None``.
+        rounding (int, str or None):
+            Define rounding scheme for data. If set to an int, values will be rounded
+            to that number of decimal places. If ``None``, values will not be rounded.
+            If set to ``'auto'``, the transformer will round to the maximum number of
+            decimal places detected in the fitted data.
+        min_value (int, str or None):
+            Indicate whether or not to set a minimum value for the data. If an integer is given,
+            reverse transformed data will be greater than or equal to it. If the string ``'auto'``
+            is given, the minimum will be the minimum value seen in the fitted data. If ``None``
+            is given, there won't be a minimum.
+        max_value (int, str or None):
+            Indicate whether or not to set a maximum value for the data. If an integer is given,
+            reverse transformed data will be less than or equal to it. If the string ``'auto'``
+            is given, the maximum will be the maximum value seen in the fitted data. If ``None``
+            is given, there won't be a maximum.
     """
 
     null_transformer = None
+    nan = None
+    _dtype = None
+    _rounding_digits = None
+    _min_value = None
+    _max_value = None
 
-    def __init__(self, dtype=None, nan='mean', null_column=None):
+    def __init__(self, dtype=None, nan='mean', null_column=None, rounding=None,
+                 min_value=None, max_value=None):
         self.nan = nan
         self.null_column = null_column
         self.dtype = dtype
-        self._dtype = dtype
+        self.rounding = rounding
+        self.min_value = min_value
+        self.max_value = max_value
+
+    @staticmethod
+    def _learn_rounding_digits(data):
+        # check if data has any decimals
+        roundable_data = data[~(np.isinf(data) | pd.isnull(data))]
+        if (roundable_data % 1 != 0).any():
+            if not (roundable_data == roundable_data.round(MAX_DECIMALS)).all():
+                return None
+
+            for decimal in range(MAX_DECIMALS + 1):
+                if (roundable_data == roundable_data.round(decimal)).all():
+                    return decimal
+
+        else:
+            maximum = max(abs(roundable_data))
+            start = int(np.log10(maximum)) if maximum != 0 else 0
+            for decimal in range(-start, 1):
+                if (roundable_data == roundable_data.round(decimal)).all():
+                    return decimal
+
+        return None
 
     def fit(self, data):
         """Fit the transformer to the data.
@@ -56,6 +102,14 @@ class NumericalTransformer(BaseTransformer):
             data = pd.Series(data)
 
         self._dtype = self.dtype or data.dtype
+        self._min_value = data.min() if self.min_value == 'auto' else self.min_value
+        self._max_value = data.max() if self.max_value == 'auto' else self.max_value
+
+        if self.rounding == 'auto':
+            self._rounding_digits = self._learn_rounding_digits(data)
+        elif isinstance(self.rounding, int):
+            self._rounding_digits = self.rounding
+
         self.null_transformer = NullTransformer(self.nan, self.null_column, copy=True)
         self.null_transformer.fit(data)
 
@@ -87,14 +141,20 @@ class NumericalTransformer(BaseTransformer):
         Returns:
             numpy.ndarray
         """
+        if self._min_value is not None or self._max_value is not None:
+            if len(data.shape) > 1:
+                data[:, 0] = data[:, 0].clip(self._min_value, self._max_value)
+            else:
+                data = data.clip(self._min_value, self._max_value)
+
         if self.nan is not None:
             data = self.null_transformer.reverse_transform(data)
 
-        if np.dtype(self._dtype).kind == 'i':
-            if pd.notnull(data).all():
-                return data.round().astype(self._dtype)
+        is_integer = np.dtype(self._dtype).kind == 'i'
+        if self._rounding_digits is not None or is_integer:
+            data = data.round(self._rounding_digits or 0)
 
-            data[pd.notnull(data)] = np.round(data[pd.notnull(data)]).astype(self._dtype)
+        if pd.isnull(data).any() and is_integer:
             return data
 
         return data.astype(self._dtype)
@@ -161,64 +221,77 @@ class GaussianCopulaTransformer(NumericalTransformer):
                 * ``truncated_gaussian``: Use a Truncated Gaussian distribution.
     """
 
-    _DISTRIBUTIONS = {
-        'univariate': univariate.Univariate,
-        'parametric': (
-            univariate.Univariate, {
-                'parametric': univariate.ParametricType.PARAMETRIC,
-            },
-        ),
-        'bounded': (
-            univariate.Univariate,
-            {
-                'bounded': univariate.BoundedType.BOUNDED,
-            },
-        ),
-        'semi_bounded': (
-            univariate.Univariate,
-            {
-                'bounded': univariate.BoundedType.SEMI_BOUNDED,
-            },
-        ),
-        'parametric_bounded': (
-            univariate.Univariate,
-            {
-                'parametric': univariate.ParametricType.PARAMETRIC,
-                'bounded': univariate.BoundedType.BOUNDED,
-            },
-        ),
-        'parametric_semi_bounded': (
-            univariate.Univariate,
-            {
-                'parametric': univariate.ParametricType.PARAMETRIC,
-                'bounded': univariate.BoundedType.SEMI_BOUNDED,
-            },
-        ),
-        'gaussian': univariate.GaussianUnivariate,
-        'gamma': univariate.GammaUnivariate,
-        'beta': univariate.BetaUnivariate,
-        'student_t': univariate.StudentTUnivariate,
-        'gaussian_kde': univariate.GaussianKDE,
-        'truncated_gaussian': univariate.TruncatedGaussian,
-    }
-
     _univariate = None
 
     def __init__(self, dtype=None, nan='mean', null_column=None, distribution='parametric'):
         super().__init__(dtype=dtype, nan=nan, null_column=null_column)
+        self._distributions = self._get_distributions()
 
         if isinstance(distribution, str):
-            distribution = self._DISTRIBUTIONS[distribution]
+            distribution = self._distributions[distribution]
 
         self._distribution = distribution
 
+    @staticmethod
+    def _get_distributions():
+        try:
+            from copulas import univariate  # pylint: disable=import-outside-toplevel
+        except ImportError as error:
+            error.msg += (
+                '\n\nIt seems like `copulas` is not installed.\n'
+                'Please install it using:\n\n    pip install rdt[copulas]'
+            )
+            raise
+
+        return {
+            'univariate': univariate.Univariate,
+            'parametric': (
+                univariate.Univariate, {
+                    'parametric': univariate.ParametricType.PARAMETRIC,
+                },
+            ),
+            'bounded': (
+                univariate.Univariate,
+                {
+                    'bounded': univariate.BoundedType.BOUNDED,
+                },
+            ),
+            'semi_bounded': (
+                univariate.Univariate,
+                {
+                    'bounded': univariate.BoundedType.SEMI_BOUNDED,
+                },
+            ),
+            'parametric_bounded': (
+                univariate.Univariate,
+                {
+                    'parametric': univariate.ParametricType.PARAMETRIC,
+                    'bounded': univariate.BoundedType.BOUNDED,
+                },
+            ),
+            'parametric_semi_bounded': (
+                univariate.Univariate,
+                {
+                    'parametric': univariate.ParametricType.PARAMETRIC,
+                    'bounded': univariate.BoundedType.SEMI_BOUNDED,
+                },
+            ),
+            'gaussian': univariate.GaussianUnivariate,
+            'gamma': univariate.GammaUnivariate,
+            'beta': univariate.BetaUnivariate,
+            'student_t': univariate.StudentTUnivariate,
+            'gaussian_kde': univariate.GaussianKDE,
+            'truncated_gaussian': univariate.TruncatedGaussian,
+        }
+
     def _get_univariate(self):
         distribution = self._distribution
-        if isinstance(distribution, univariate.Univariate):
+        if isinstance(distribution, self._distributions['univariate']):
             return copy.deepcopy(distribution)
         if isinstance(distribution, tuple):
             return distribution[0](**distribution[1])
-        if isinstance(distribution, type) and issubclass(distribution, univariate.Univariate):
+        if isinstance(distribution, type) and \
+           issubclass(distribution, self._distributions['univariate']):
             return distribution()
 
         raise TypeError('Invalid distribution: {}'.format(distribution))
